@@ -109,8 +109,9 @@ func newCluster(t *testing.T, objs ...runtime.Object) *clusterCache {
 
 func getChildren(cluster *clusterCache, un *unstructured.Unstructured) []*Resource {
 	hierarchy := make([]*Resource, 0)
-	cluster.IterateHierarchy(kube.GetResourceKey(un), func(child *Resource, _ map[kube.ResourceKey]*Resource) {
+	cluster.IterateHierarchy(kube.GetResourceKey(un), func(child *Resource, _ map[kube.ResourceKey]*Resource) bool {
 		hierarchy = append(hierarchy, child)
+		return true
 	})
 	return hierarchy[1:]
 }
@@ -620,16 +621,54 @@ func TestGetDuplicatedChildren(t *testing.T) {
 
 func TestGetClusterInfo(t *testing.T) {
 	cluster := newCluster(t)
-	cluster.apiGroups = []metav1.APIGroup{{Name: "test"}}
+	cluster.apiResources = []kube.APIResourceInfo{{GroupKind: schema.GroupKind{Group: "test", Kind: "test kind"}}}
 	cluster.serverVersion = "v1.16"
 	info := cluster.GetClusterInfo()
 	assert.Equal(t, ClusterInfo{
-		Server:     cluster.config.Host,
-		APIGroups:  cluster.apiGroups,
-		K8SVersion: cluster.serverVersion,
+		Server:       cluster.config.Host,
+		APIResources: cluster.apiResources,
+		K8SVersion:   cluster.serverVersion,
 	}, info)
 }
 
+func TestDeleteAPIResource(t *testing.T) {
+	cluster := newCluster(t)
+	cluster.apiResources = []kube.APIResourceInfo{{
+		GroupKind:            schema.GroupKind{Group: "test", Kind: "test kind"},
+		GroupVersionResource: schema.GroupVersionResource{Version: "v1"},
+	}}
+
+	cluster.deleteAPIResource(kube.APIResourceInfo{GroupKind: schema.GroupKind{Group: "wrong group", Kind: "wrong kind"}})
+	assert.Len(t, cluster.apiResources, 1)
+	cluster.deleteAPIResource(kube.APIResourceInfo{
+		GroupKind:            schema.GroupKind{Group: "test", Kind: "test kind"},
+		GroupVersionResource: schema.GroupVersionResource{Version: "wrong version"},
+	})
+	assert.Len(t, cluster.apiResources, 1)
+
+	cluster.deleteAPIResource(kube.APIResourceInfo{
+		GroupKind:            schema.GroupKind{Group: "test", Kind: "test kind"},
+		GroupVersionResource: schema.GroupVersionResource{Version: "v1"},
+	})
+	assert.Empty(t, cluster.apiResources)
+}
+
+func TestAppendAPIResource(t *testing.T) {
+	cluster := newCluster(t)
+
+	resourceInfo := kube.APIResourceInfo{
+		GroupKind:            schema.GroupKind{Group: "test", Kind: "test kind"},
+		GroupVersionResource: schema.GroupVersionResource{Version: "v1"},
+	}
+
+	cluster.appendAPIResource(resourceInfo)
+	assert.ElementsMatch(t, []kube.APIResourceInfo{resourceInfo}, cluster.apiResources)
+
+	// make sure same group, kind version is not added twice
+	cluster.appendAPIResource(resourceInfo)
+	assert.ElementsMatch(t, []kube.APIResourceInfo{resourceInfo}, cluster.apiResources)
+}
+
 func ExampleNewClusterCache_resourceUpdatedEvents() {
 	// kubernetes cluster config here
 	config := &rest.Config{}
@@ -758,3 +797,68 @@ func testDeploy() *appsv1.Deployment {
 		},
 	}
 }
+
+func TestIterateHierachy(t *testing.T) {
+	cluster := newCluster(t, testPod(), testRS(), testDeploy())
+	err := cluster.EnsureSynced()
+	require.NoError(t, err)
+
+	t.Run("IterateAll", func(t *testing.T) {
+		keys := []kube.ResourceKey{}
+		cluster.IterateHierarchy(kube.GetResourceKey(mustToUnstructured(testDeploy())), func(child *Resource, _ map[kube.ResourceKey]*Resource) bool {
+			keys = append(keys, child.ResourceKey())
+			return true
+		})
+
+		assert.ElementsMatch(t,
+			[]kube.ResourceKey{
+				kube.GetResourceKey(mustToUnstructured(testPod())),
+				kube.GetResourceKey(mustToUnstructured(testRS())),
+				kube.GetResourceKey(mustToUnstructured(testDeploy()))},
+			keys)
+	})
+
+	t.Run("ExitAtRoot", func(t *testing.T) {
+		keys := []kube.ResourceKey{}
+		cluster.IterateHierarchy(kube.GetResourceKey(mustToUnstructured(testDeploy())), func(child *Resource, _ map[kube.ResourceKey]*Resource) bool {
+			keys = append(keys, child.ResourceKey())
+			return false
+		})
+
+		assert.ElementsMatch(t,
+			[]kube.ResourceKey{
+				kube.GetResourceKey(mustToUnstructured(testDeploy()))},
+			keys)
+	})
+
+	t.Run("ExitAtSecondLevelChild", func(t *testing.T) {
+		keys := []kube.ResourceKey{}
+		cluster.IterateHierarchy(kube.GetResourceKey(mustToUnstructured(testDeploy())), func(child *Resource, _ map[kube.ResourceKey]*Resource) bool {
+			keys = append(keys, child.ResourceKey())
+			return child.ResourceKey().Kind != kube.ReplicaSetKind
+		})
+
+		assert.ElementsMatch(t,
+			[]kube.ResourceKey{
+				kube.GetResourceKey(mustToUnstructured(testDeploy())),
+				kube.GetResourceKey(mustToUnstructured(testRS())),
+			},
+			keys)
+	})
+
+	t.Run("ExitAtThirdLevelChild", func(t *testing.T) {
+		keys := []kube.ResourceKey{}
+		cluster.IterateHierarchy(kube.GetResourceKey(mustToUnstructured(testDeploy())), func(child *Resource, _ map[kube.ResourceKey]*Resource) bool {
+			keys = append(keys, child.ResourceKey())
+			return child.ResourceKey().Kind != kube.PodKind
+		})
+
+		assert.ElementsMatch(t,
+			[]kube.ResourceKey{
+				kube.GetResourceKey(mustToUnstructured(testDeploy())),
+				kube.GetResourceKey(mustToUnstructured(testRS())),
+				kube.GetResourceKey(mustToUnstructured(testPod())),
+			},
+			keys)
+	})
+}
