@@ -12,7 +12,7 @@ import (
 
 	"github.com/go-logr/logr"
 	v1 "k8s.io/api/core/v1"
-	"k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1beta1"
+	v1extensions "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1"
 	"k8s.io/apiextensions-apiserver/pkg/client/clientset/clientset"
 	apierr "k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
@@ -23,6 +23,7 @@ import (
 	"k8s.io/client-go/rest"
 	"k8s.io/klog/v2/klogr"
 	cmdutil "k8s.io/kubectl/pkg/cmd/util"
+	"k8s.io/kubectl/pkg/util/openapi"
 
 	"github.com/argoproj/gitops-engine/pkg/diff"
 	"github.com/argoproj/gitops-engine/pkg/health"
@@ -178,7 +179,7 @@ func WithReplace(replace bool) SyncOpt {
 	}
 }
 
-//  NewSyncContext creates new instance of a SyncContext
+// NewSyncContext creates new instance of a SyncContext
 func NewSyncContext(
 	revision string,
 	reconciliationResult ReconciliationResult,
@@ -186,19 +187,24 @@ func NewSyncContext(
 	rawConfig *rest.Config,
 	kubectl kubeutil.Kubectl,
 	namespace string,
+	openAPISchema openapi.Resources,
 	opts ...SyncOpt,
-) (SyncContext, error) {
+) (SyncContext, func(), error) {
 	dynamicIf, err := dynamic.NewForConfig(restConfig)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 	disco, err := discovery.NewDiscoveryClientForConfig(restConfig)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
 	}
 	extensionsclientset, err := clientset.NewForConfig(restConfig)
 	if err != nil {
-		return nil, err
+		return nil, nil, err
+	}
+	resourceOps, cleanup, err := kubectl.ManageResources(rawConfig, openAPISchema)
+	if err != nil {
+		return nil, nil, err
 	}
 	ctx := &syncContext{
 		revision:            revision,
@@ -210,6 +216,7 @@ func NewSyncContext(
 		disco:               disco,
 		extensionsclientset: extensionsclientset,
 		kubectl:             kubectl,
+		resourceOps:         resourceOps,
 		namespace:           namespace,
 		log:                 klogr.New(),
 		validate:            true,
@@ -222,7 +229,7 @@ func NewSyncContext(
 	for _, opt := range opts {
 		opt(ctx)
 	}
-	return ctx, nil
+	return ctx, cleanup, nil
 }
 
 func groupResources(reconciliationResult ReconciliationResult) map[kubeutil.ResourceKey]reconciledResource {
@@ -303,6 +310,7 @@ type syncContext struct {
 	disco               discovery.DiscoveryInterface
 	extensionsclientset *clientset.Clientset
 	kubectl             kube.Kubectl
+	resourceOps         kube.ResourceOperations
 	namespace           string
 
 	dryRun                 bool
@@ -442,10 +450,12 @@ func (sc *syncContext) Sync() {
 	// syncFailTasks only run during failure, so separate them from regular tasks
 	syncFailTasks, tasks := tasks.Split(func(t *syncTask) bool { return t.phase == common.SyncPhaseSyncFail })
 
+	syncFailedTasks, _ := tasks.Split(func(t *syncTask) bool { return t.syncStatus == common.ResultCodeSyncFailed })
+
 	// if there are any completed but unsuccessful tasks, sync is a failure.
 	if tasks.Any(func(t *syncTask) bool { return t.completed() && !t.successful() }) {
 		sc.deleteHooks(hooksPendingDeletionFailed)
-		sc.setOperationFailed(syncFailTasks, "one or more synchronization tasks completed unsuccessfully")
+		sc.setOperationFailed(syncFailTasks, syncFailedTasks, "one or more synchronization tasks completed unsuccessfully")
 		return
 	}
 
@@ -496,8 +506,9 @@ func (sc *syncContext) Sync() {
 
 	switch runState {
 	case failed:
+		syncFailedTasks, _ := tasks.Split(func(t *syncTask) bool { return t.syncStatus == common.ResultCodeSyncFailed })
 		sc.deleteHooks(hooksPendingDeletionFailed)
-		sc.setOperationFailed(syncFailTasks, "one or more objects failed to apply")
+		sc.setOperationFailed(syncFailTasks, syncFailedTasks, "one or more objects failed to apply")
 	case successful:
 		if remainingTasks.Len() == 0 {
 			// delete all completed hooks which have appropriate delete policy
@@ -548,21 +559,33 @@ func (sc *syncContext) GetState() (common.OperationPhase, string, []common.Resou
 	return sc.phase, sc.message, resourceRes
 }
 
-func (sc *syncContext) setOperationFailed(syncFailTasks syncTasks, message string) {
+func (sc *syncContext) setOperationFailed(syncFailTasks, syncFailedTasks syncTasks, message string) {
+	errorMessageFactory := func(tasks []*syncTask, message string) string {
+		messages := syncFailedTasks.Map(func(task *syncTask) string {
+			return task.message
+		})
+		if len(messages) > 0 {
+			return fmt.Sprintf("%s, reason: %s", message, strings.Join(messages, ","))
+		}
+		return message
+	}
+
+	errorMessage := errorMessageFactory(syncFailedTasks, message)
+
 	if len(syncFailTasks) > 0 {
 		// if all the failure hooks are completed, don't run them again, and mark the sync as failed
 		if syncFailTasks.All(func(task *syncTask) bool { return task.completed() }) {
-			sc.setOperationPhase(common.OperationFailed, message)
+			sc.setOperationPhase(common.OperationFailed, errorMessage)
 			return
 		}
 		// otherwise, we need to start the failure hooks, and then return without setting
 		// the phase, so we make sure we have at least one more sync
 		sc.log.WithValues("syncFailTasks", syncFailTasks).V(1).Info("Running sync fail tasks")
 		if sc.runTasks(syncFailTasks, false) == failed {
-			sc.setOperationPhase(common.OperationFailed, message)
+			sc.setOperationPhase(common.OperationFailed, errorMessage)
 		}
 	} else {
-		sc.setOperationPhase(common.OperationFailed, message)
+		sc.setOperationPhase(common.OperationFailed, errorMessage)
 	}
 }
 
@@ -775,7 +798,6 @@ func isNamespaceWithName(res *unstructured.Unstructured, ns string) bool {
 
 func isNamespaceKind(res *unstructured.Unstructured) bool {
 	return res != nil &&
-		res.GetObjectKind() != nil &&
 		res.GetObjectKind().GroupVersionKind().Group == "" &&
 		res.GetKind() == kube.NamespaceKind
 }
@@ -809,16 +831,16 @@ func (sc *syncContext) setOperationPhase(phase common.OperationPhase, message st
 	sc.message = message
 }
 
-// ensureCRDReady waits until specified CRD is ready (established condition is true). Method is best effort - it does not fail even if CRD is not ready without timeout.
-func (sc *syncContext) ensureCRDReady(name string) {
-	_ = wait.PollImmediate(time.Duration(100)*time.Millisecond, crdReadinessTimeout, func() (bool, error) {
-		crd, err := sc.extensionsclientset.ApiextensionsV1beta1().CustomResourceDefinitions().Get(context.TODO(), name, metav1.GetOptions{})
+// ensureCRDReady waits until specified CRD is ready (established condition is true).
+func (sc *syncContext) ensureCRDReady(name string) error {
+	return wait.PollImmediate(time.Duration(100)*time.Millisecond, crdReadinessTimeout, func() (bool, error) {
+		crd, err := sc.extensionsclientset.ApiextensionsV1().CustomResourceDefinitions().Get(context.TODO(), name, metav1.GetOptions{})
 		if err != nil {
 			return false, err
 		}
 		for _, condition := range crd.Status.Conditions {
-			if condition.Type == v1beta1.Established {
-				return condition.Status == v1beta1.ConditionTrue, nil
+			if condition.Type == v1extensions.Established {
+				return condition.Status == v1extensions.ConditionTrue, nil
 			}
 		}
 		return false, nil
@@ -840,31 +862,29 @@ func (sc *syncContext) applyObject(t *syncTask, dryRun bool, force bool, validat
 			if kube.IsCRD(t.targetObj) {
 				update := t.targetObj.DeepCopy()
 				update.SetResourceVersion(t.liveObj.GetResourceVersion())
-				_, err = sc.kubectl.UpdateResource(context.TODO(), sc.rawConfig, update, t.targetObj.GetNamespace(), dryRunStrategy)
+				_, err = sc.resourceOps.UpdateResource(context.TODO(), update, dryRunStrategy)
 				if err == nil {
 					message = fmt.Sprintf("%s/%s updated", t.targetObj.GetKind(), t.targetObj.GetName())
 				} else {
 					message = fmt.Sprintf("error when updating: %v", err.Error())
 				}
 			} else {
-				message, err = sc.kubectl.ReplaceResource(context.TODO(), sc.rawConfig, t.targetObj, t.targetObj.GetNamespace(), dryRunStrategy, force)
+				message, err = sc.resourceOps.ReplaceResource(context.TODO(), t.targetObj, dryRunStrategy, force)
 			}
 		} else {
-			_, err = sc.kubectl.CreateResource(context.TODO(), sc.rawConfig, t.targetObj, t.targetObj.GetNamespace(), dryRunStrategy)
-			if err == nil {
-				message = fmt.Sprintf("%s/%s created", t.targetObj.GetKind(), t.targetObj.GetName())
-			} else {
-				message = fmt.Sprintf("error when creating: %v", err.Error())
-			}
+			message, err = sc.resourceOps.CreateResource(context.TODO(), t.targetObj, dryRunStrategy, validate)
 		}
 	} else {
-		message, err = sc.kubectl.ApplyResource(context.TODO(), sc.rawConfig, t.targetObj, t.targetObj.GetNamespace(), dryRunStrategy, force, validate)
+		message, err = sc.resourceOps.ApplyResource(context.TODO(), t.targetObj, dryRunStrategy, force, validate)
 	}
 	if err != nil {
 		return common.ResultCodeSyncFailed, err.Error()
 	}
 	if kube.IsCRD(t.targetObj) && !dryRun {
-		sc.ensureCRDReady(t.targetObj.GetName())
+		crdName := t.targetObj.GetName()
+		if err = sc.ensureCRDReady(crdName); err != nil {
+			sc.log.Error(err, fmt.Sprintf("failed to ensure that CRD %s is ready", crdName))
+		}
 	}
 	return common.ResultCodeSynced, message
 }
