@@ -1,6 +1,6 @@
 import * as moment from 'moment';
 import * as React from 'react';
-import {Observable, Subscription} from 'rxjs';
+import {interval, Subscription} from 'rxjs';
 
 export class Ticker extends React.Component<{intervalMs?: number, disabled?: boolean, children?: ((time: moment.Moment) => React.ReactNode)}, {time: moment.Moment}> {
 
@@ -28,7 +28,7 @@ export class Ticker extends React.Component<{intervalMs?: number, disabled?: boo
         if (this.props.disabled) {
             this.ensureUnsubscribed();
         } else if (!this.subscription) {
-            this.subscription = Observable.interval(this.props.intervalMs || 1000).subscribe(() => this.setState({ time: moment() }));
+            this.subscription = interval(this.props.intervalMs || 1000).subscribe(() => this.setState({ time: moment() }));
         }
     }
 
